# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Derived class of server functions for Numpy/Scipy operations

    take a stream of assets which are raw acceleration values, process them, and return
    a stream of assets representing the results.

"""


__author__ = "Douglas Orr, Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import logging
import datetime

import scipy.signal as sig
from scipy import fftpack
import numpy as np

from foglamp.common import logger

G_ACCEL = 9.80665

_LOGGER = logger.setup(__name__, level=logging.INFO)


def bandwidth_filter_high(data, cutoff, order=10):
    """ High pass filter
    Args:
        data: signal data
        cutoff: cutoff frequency divided by Nyquist frequency
        order: Order of the butterworth filter
    Returns:
        filtered signal
    Raises:
    """
    _sos = sig.butter(order, cutoff, btype='high', output='sos')
    return sig.sosfiltfilt(_sos, data)


def bandwidth_filter_low(data, cutoff, order=10):
    """ Low pass filter
    Args:
        data: signal data
        cutoff: cutoff frequency divided by Nyquist frequency
        order: Order of the butterworth filter
    Returns:
        filtered signal
    Raises:
    """
    _sos = sig.butter(order, cutoff, btype='low', output='sos')
    return sig.sosfiltfilt(_sos, data)


def bandwidth_filter_bandpass(data, low_high_cutoff, order=10):
    """ Band pass filter
    Args:
        data: signal data
        low_high_cutoff: 2 element list with low & high cutoff frequency divided by Nyquist frequency
        order: Order of the butterworth filter
    Returns:
        filtered signal
    Raises:
    """
    _sos = sig.butter(order, low_high_cutoff, btype='bandpass', output='sos')
    return sig.sosfiltfilt(_sos, data)


def filter_signal(data, sampling_rate, f_lower, f_upper):
    """ Filters signal
    Args:
        data: signal data
        sampling_rate: sampling rate of signal
        f_lower: lower frequency of the filter
        f_upper: upper frequency of the filter
    Returns:
        filtered signal
    Raises:
    """
    nyq = sampling_rate / 2.0

    f_lower = max(0, f_lower)
    f_upper = min(sampling_rate/2, f_upper)

    if (f_lower == 0) and (f_upper == sampling_rate/2):
        data = data
    elif f_lower == 0:
        data = bandwidth_filter_low(data, f_upper/nyq)
    elif f_upper == sampling_rate/2:
        data = bandwidth_filter_high(data, f_lower/nyq)
    else:
        data = bandwidth_filter_bandpass(data, [f_lower/nyq, f_upper/nyq])

    return data


def hilbert(data):
    """ calculate absolute value of hilbert transform of signal
    Args:
        data: signal data
    Returns:
        absolute value of hilbert transform of signal
    Raises:
    """
    analytic_signal = sig.hilbert(data)
    return np.abs(analytic_signal)


def acc_to_velocity(signal, fs, calc_envelope=True, f_lower=4500, f_upper=8000):
    """ Processes raw acceleration data to calculate velocity, RMS acceleration/velocity & acceleration envelope
    Args:
        signal: signal data
        fs: sampling frequency of the signal
        calc_envelope: whether to calculate acceleration envelope
        f_lower,f_upper: frequency range of pass band to use during envelope calculation
    Returns:
        velocity: instantaneous velocity
        accelGRms: RMS acceleration in G
        accelGRmsPeak: Peak acceleration in G
        velGRms: RMS velocity in mm/s
        velGRmsPeak: Peak velocity in mm/s
        envelope: Acceleration envelope
    Raises:
    """
    t1 = datetime.datetime.now()

    signal = np.array(signal)
    _LOGGER.debug("signal.shape={}, signal[0]={}".format(signal.shape, signal[0]))

    accelGRms = list()
    accelGRmsPeak = list()
    velocity = list()
    velGRms = list()
    velGRmsPeak = list()

    for col in range(signal.shape[1]):
        _sig = signal[:, col]
        _sig = sig.detrend(_sig)
        _LOGGER.debug("col={}, _sig.shape={}, _sig[:5]={}".format(col, _sig.shape, _sig[:5]))
        accelGRms.append(np.sqrt(np.mean(_sig ** 2)) / G_ACCEL)
        accelGRmsPeak.append(np.max(_sig) / G_ACCEL)

        amps = fftpack.fft(_sig)
        _LOGGER.debug("amps.shape={}, amps[0]={}".format(amps.shape, amps[0]))
        freqs = fftpack.fftfreq(_sig.shape[0], d=1/fs)
        _LOGGER.debug("freqs.shape={}, freqs[0]={}".format(freqs.shape, freqs[0]))
        fft_bin_width = freqs[1]
        skip_bins = int(round((fs/1000)/fft_bin_width))

        amps = amps / (2 * np.pi * (1j) * (freqs + [fft_bin_width/2]))
        _LOGGER.debug("amps.shape={}, amps[0]={}".format(amps.shape, amps[0]))
        amps[:skip_bins] = [0]
        amps[-skip_bins:] = [0]
        _LOGGER.debug("amps.shape={}, amps[0]={}".format(amps.shape, amps[0]))

        vel = fftpack.ifft(amps)
        _LOGGER.debug("vel.shape={}, vel[0]={}".format(vel.shape, vel[0]))

        vel = np.real(vel)
        vel = vel * [1000]  # convert from m/s to mm/s
        # _LOGGER.info("Step 1")
        velocity.append(vel)
        # _LOGGER.info("Step 2")

        velGRms.append(np.sqrt(np.mean(vel ** 2)))
        # _LOGGER.info("Step 3")
        # _LOGGER.info("velGRms.shape={}, velGRms[0]={}".format(velGRms.shape, velGRms[0]))
        velGRmsPeak.append(np.max(vel))  # np.max(np.abs(signal))
        # _LOGGER.info("Step 4")
        # _LOGGER.info("velGRmsPeak.shape={}, velGRmsPeak[0]={}".format(velGRmsPeak.shape, velGRmsPeak[0]))

        if calc_envelope:
            # _LOGGER.info("Step 5")
            envelope = hilbert(filter_signal(_sig, fs, f_lower, f_upper))
            envelope = envelope / [G_ACCEL]
            envelope = envelope.tolist()
        else:
            # _LOGGER.info("Step 6")
            envelope = None

    _LOGGER.debug("len(velocity)={}, type(velocity[0])={}, velocity[0].shape={}"
                 .format(len(velocity), type(velocity[0]), velocity[0].shape))
    # _LOGGER.info("Step 7")
    velocity = np.array(velocity).T  # back to same shape as input signal
    _LOGGER.debug("velocity.shape={}, velocity[0]={}".format(velocity.shape, velocity[0]))

    t2 = datetime.datetime.now()
    delta = t2 - t1
    _LOGGER.debug("acc_to_velocity took {} usec".format(int(delta.total_seconds() * 1000 * 1000)))

    return velocity, np.array(accelGRms), np.array(accelGRmsPeak), np.array(velGRms), np.array(velGRmsPeak), envelope
