# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for velocity filter plugin
This filter calculates velocity values from acceleration data at the original sampling rate. It also adds a datapoint
for storing acceleration value in G's. It also adds a new reading every second containing RMS and peak values for
acceleration and velocity every second.
"""

import os
import time
import copy
import logging

from foglamp.common import logger
from foglamp.plugins.common import utils
import filter_ingest
from foglamp.plugins.filter.velocity.np_calculations import acc_to_velocity

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
_LOGGER = logger.setup(__name__, level=logging.INFO)

_module_dir = os.path.dirname(__file__)

G_ACCEL = 9.80665

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Velocity filter plugin',
        'type': 'string',
        'default': 'velocity',
        'readonly': 'true'
    },
    'accelAssetName': {
        'description': 'Input asset name of reading with acceleration value',
        'type': 'string',
        'default': 'vibration',
        'displayName': 'Acceleration asset name',
        'order': "1"
    },
    'accelDatapointNames': {
        'description': 'Comma separated list of Datapoint names in reading holding the acceleration value (in m/s^2)',
        'type': 'string',
        'default': '20gpk_accelmss',
        'displayName': 'Acceleration datapoint names',
        'order': "2"
    },
    'sampling_rate': {
        'description': 'Sampling rate of acceleration data (Hz)',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sampling rate (samples per sec)',
        'order': "3"
    },
    'updateOriginalReading': {
        'description': 'Whether to update each original reading with instantaneous velocity and acceleration (in G) values',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Update readings with instantaneous vel/accel-G',
        'order': "4"
    },
    'accelGDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold acceleration value in G (empty string to exclude this datapoint)',
        'type': 'string',
        'default': '_accelG',
        'displayName': 'Acceleration-G datapoint name suffix',
        'validity': "updateOriginalReading == 'true'",
        'order': "5"
    },
    'velDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold the velocity value in mm/s (empty string to exclude this datapoint)',
        'type': 'string',
        'default': '_velocity_mms',
        'displayName': 'Velocity datapoint name suffix',
        'validity': "updateOriginalReading == 'true'",
        'order': "6"
    },
    'numSamplesForIntegration': {
        'description': 'Number of data points to use for integration and RMS calculation',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Number of samples to use for integration',
        'minimum': '3',
        'order': "7"
    },
    'enableRmsFeatures': {
        'description': 'Whether to enable RMS features',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enable RMS features',
        'order': "8"
    },
    'accelGRmsDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold RMS acceleration value in G (empty string to exclude this datapoint)',
        'type': 'string',
        'default': '_accelGRms',
        'displayName': 'Acceleration-G RMS datapoint name suffix',
        'validity': "enableRmsFeatures == 'true'",
        'order': "9"
    },
    'accelGRmsPeakDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold RMS acceleration peak value in G (empty string to exclude this datapoint)',
        'type': 'string',
        'default': '_accelGRms_pk',
        'displayName': 'Acceleration-G RMS peak datapoint name suffix',
        'validity': "enableRmsFeatures == 'true'",
        'order': "10"
    },
    'velRmsDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold the RMS velocity value in mm/s (empty string to exclude this datapoint)',
        'type': 'string',
        'default': '_velocityRms',
        'displayName': 'Velocity RMS datapoint name suffix',
        'validity': "enableRmsFeatures == 'true'",
        'order': "11"
    },
    'velRmsPeakDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold the RMS velocity peak value in mm/s (empty string to exclude this datapoint)',
        'type': 'string',
        'default': '_velocityRms_pk',
        'displayName': 'Velocity RMS peak datapoint name suffix',
        'validity': "enableRmsFeatures == 'true'",
        'order': "12"
    },
    'enableEnvelope': {
        'description': 'Whether to enable acceleration envelope output',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enable envelope output',
        'order': "13"
    },
    'accelGEnvelopeDatapointNameSuffix': {
        'description': 'Suffix for datapoint that would hold acceleration envelope value in G',
        'type': 'string',
        'default': '_accelG_envelope',
        'displayName': 'Acceleration-G envelope datapoint name suffix',
        'validity': "enableEnvelope == 'true'",
        'order': "14"
    },
    'resonanceFreqLower': {
        'description': 'Resonance band start frequency of bearing for envelope calculation (Hz)',
        'type': 'integer',
        'default': '1000',
        'displayName': 'Bearing resonance band start frequency (Hz)',
        'minimum': '0',
        'validity': "enableEnvelope == 'true'",
        'order': "15",
    },
    'resonanceFreqUpper': {
        'description': 'Resonance band end frequency of bearing for envelope calculation (Hz)',
        'type': 'integer',
        'default': '4000',
        'displayName': 'Bearing resonance band end frequency (Hz)',
        'minimum': '0',
        'validity': "enableEnvelope == 'true'",
        'order': "16"
    },
    'enable': {
        'description': 'Enable filter plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "17"
    },
}


def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug("plugin_info called")
    return {
        'name': 'velocity',
        'version': '2.1.0',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def parse_config(handle, config):
    """ Initialise the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        config: JSON configuration document for the Filter plugin configuration category
    Returns:
        data: updated handle
    Raises:
    """
    handle['accelAssetName'] = config['accelAssetName']['value']
    handle['accelDatapointNames'] = config['accelDatapointNames']['value'].split(',')
    handle['sampling_rate'] = int(config['sampling_rate']['value'])
    handle['enable'] = config['enable']['value'] == 'true'

    handle['updateOriginalReading'] = config['updateOriginalReading']['value'] == 'true'
    handle['accelGDatapointNameSuffix'] = config['accelGDatapointNameSuffix']['value']
    handle['velDatapointNameSuffix'] = config['velDatapointNameSuffix']['value']
    handle['numSamplesForIntegration'] = int(config['numSamplesForIntegration']['value'])

    handle['enableRmsFeatures'] = config['enableRmsFeatures']['value'] == 'true'
    handle['accelGRmsDatapointNameSuffix'] = config['accelGRmsDatapointNameSuffix']['value']
    handle['accelGRmsPeakDatapointNameSuffix'] = config['accelGRmsPeakDatapointNameSuffix']['value']
    handle['velRmsDatapointNameSuffix'] = config['velRmsDatapointNameSuffix']['value']
    handle['velRmsPeakDatapointNameSuffix'] = config['velRmsPeakDatapointNameSuffix']['value']

    handle['enableEnvelope'] = config['enableEnvelope']['value'] == 'true'
    handle['accelGEnvelopeDatapointNameSuffix'] = config['accelGEnvelopeDatapointNameSuffix']['value']
    handle['resonanceFreqLower'] = int(config['resonanceFreqLower']['value'])
    handle['resonanceFreqUpper'] = int(config['resonanceFreqUpper']['value'])

    handle['shutdown_in_progress'] = config['shutdown_in_progress']
    handle['callback'] = config['callback']
    handle['ingest_ref'] = config['ingest_ref']

    return handle


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref: Ingest reference object
        callback: Callback reference object
    Returns:
        handle: python dict object to be used in future calls to the plugin
    Raises:
    """
    config['shutdown_in_progress'] = False
    config['callback'] = callback
    config['ingest_ref'] = ingest_ref

    handle = dict()
    handle = parse_config(handle, config)
    handle['readings_buffer'] = list()

    return handle


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.debug("Old config for velocity plugin {} \n     new config {}".format(handle, new_config))

    new_handle = plugin_init(new_config, handle['ingest_ref'], handle['callback'])
    old_buffer = handle['readings_buffer']
    new_handle['readings_buffer'] = old_buffer
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    handle['shutdown_in_progress'] = True
    handle['callback'] = None
    handle['ingest_ref'] = None

    _LOGGER.info('waiting... for SHUTDOWN')
    time.sleep(2)
    _LOGGER.info('Filter velocity plugin shutdown.')


def plugin_ingest(handle, data):
    """ Modify readings data and pass it onward
    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """

    shutdown_in_progress = handle['shutdown_in_progress']
    the_callback = handle['callback']
    the_ingest_ref = handle['ingest_ref']

    if shutdown_in_progress:
        return

    if not handle['enable']:
        # Filter not enabled, just pass data onwards
        filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, data)
        return

    _asset_name = handle['accelAssetName']
    _datapoint_names = handle['accelDatapointNames']
    _sampling_rate = handle['sampling_rate']
    _num_samples_for_integration = handle['numSamplesForIntegration']

    if type(data) == dict:
        _LOGGER.debug("data is a DICT, data={}".format(data))
        data = [data]

    if len(data) == 0:
        _LOGGER.debug("data={}".format(data))

    buffer = list()
    handle['readings_buffer'].extend(data)
    _LOGGER.debug(
        "Got {} new readings to process, total {} in readings_buffer".format(len(data), len(handle['readings_buffer'])))
    if len(handle['readings_buffer']) > 0:
        _LOGGER.debug("Last reading received: {}".format(handle['readings_buffer'][-1]))

    if len(handle['readings_buffer']) < _num_samples_for_integration:
        _LOGGER.debug("Not processing readings_buffer yet")
        return

    for reading in handle['readings_buffer'][:_num_samples_for_integration]:
        if reading['asset'] != _asset_name:
            _LOGGER.debug("Skipping reading: reading['asset']={}, _asset_name={}".format(reading['asset'], _asset_name))
            continue

        datapoints = reading['readings']

        _all_datapoints_found = True
        for d in _datapoint_names:
            if d not in datapoints:
                _LOGGER.debug("Skipping reading, datapoint {} not in {}".format(d, list(datapoints.keys())))
                _all_datapoints_found = False
                break

        if not _all_datapoints_found:
            continue

        # accel = datapoints[_datapoint_names]
        accel = [datapoints[d] for d in _datapoint_names]
        buffer.append(accel)

    velo, accelGRms, accelGRmsPeak, velGRms, velGRmsPeak, envelope = \
        acc_to_velocity(buffer, _sampling_rate,
                        handle['enableEnvelope'],
                        handle['resonanceFreqLower'],
                        handle['resonanceFreqUpper'])

    # for arr in [velo, accelGRms, accelGRmsPeak, velGRms, velGRmsPeak, envelope]:
    # _LOGGER.debug("arr={}".format(arr))
    # if arr and len(arr):
    #     _LOGGER.debug("arr.shape={}".format(arr.shape))
    # _LOGGER.debug("velo.shape={}, accelGRms.shape={}, accelGRmsPeak.shape={}, velGRms.shape={}, velGRmsPeak.shape={}, envelope.shape={}"
    #               .format(velo.shape, accelGRms.shape, accelGRmsPeak.shape, velGRms.shape, velGRmsPeak.shape, envelope.shape))

    if handle['updateOriginalReading']:
        for i, reading in enumerate(handle['readings_buffer'][:_num_samples_for_integration]):
            for j, d in enumerate(_datapoint_names):
                if handle['velDatapointNameSuffix']:
                    reading['readings'][d + handle['velDatapointNameSuffix']] = velo[i][j]
                if handle['accelGDatapointNameSuffix']:
                    reading['readings'][d + handle['accelGDatapointNameSuffix']] = reading['readings'][d] / G_ACCEL

        _LOGGER.debug("Updated {} readings to include RMS/peak acceleration/velocity/accel-envelope values"
                      .format(_num_samples_for_integration))

    time_stamp = utils.local_timestamp()

    if handle['enableRmsFeatures']:
        rms_reading = {'asset': _asset_name + '_RMS', 'timestamp': time_stamp, 'readings': {}}

        for j, d in enumerate(_datapoint_names):
            if handle['accelGRmsDatapointNameSuffix']:
                rms_reading['readings'][d + handle['accelGRmsDatapointNameSuffix']] = accelGRms[j]
            if handle['accelGRmsPeakDatapointNameSuffix']:
                rms_reading['readings'][d + handle['accelGRmsPeakDatapointNameSuffix']] = accelGRmsPeak[j]
            if handle['velRmsDatapointNameSuffix']:
                rms_reading['readings'][d + handle['velRmsDatapointNameSuffix']] = velGRms[j]
            if handle['velRmsPeakDatapointNameSuffix']:
                rms_reading['readings'][d + handle['velRmsPeakDatapointNameSuffix']] = velGRmsPeak[j]

            # _LOGGER.debug("accelGRms={:.4f}, velGRms={:.4f}".format(accelGRms, velGRms))
        _LOGGER.debug("Last rms_reading generated: {}".format(rms_reading))

    envelope_reading = {'asset': _asset_name + '_envelope', 'timestamp': time_stamp, 'readings': {}}
    if handle['enableEnvelope']:
        for j, d in enumerate(_datapoint_names):
            envelope_reading['readings'][d + handle['accelGEnvelopeDatapointNameSuffix']] = envelope[j]

    if True:  # TODO: Make it true before committing
        readings_buffer_to_fwd = handle['readings_buffer'][:_num_samples_for_integration]
    else:
        readings_buffer_to_fwd = []  # to avoid filling disk with raw data at high rate

    if handle['enableRmsFeatures']:
        readings_buffer_to_fwd.append(rms_reading)

    if handle['enableEnvelope']:
        readings_buffer_to_fwd.append(envelope_reading)

    filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, readings_buffer_to_fwd)
    handle['readings_buffer'] = handle['readings_buffer'][_num_samples_for_integration:]
